# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Useful functions for processing images. Currently its contents are focused
on image stabilization.

Functions
------------
.. autosummary::
   :toctree: generated/

   to8bit
   SIFTRegisterTransform
   ORBRegisterTransform
   edgeDetectRegisterTranslation

"""
from __future__ import annotations
import logging
import typing
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from skimage import feature
from skimage import registration
if typing.TYPE_CHECKING:
    import cv2

logger = logging.getLogger(__name__)


def to8bit(arr: np.ndarray) -> np.ndarray:
    """Converts boolean or float type numpy arrays to 8bit and scales the data to span from 0 to 255. Used for many
    OpenCV functions.

    Args:
        arr: The input array

    Returns:
        The output array of dtype numpy.uint8
    """
    if arr.dtype == bool:
        arr = arr.astype(np.uint8) * 255
    else:
        arr = arr.astype(float) # This solves problems if the array is int16 type
        Min = np.percentile(arr, 0.1)
        arr -= Min
        Max = np.percentile(arr, 99.9)
        arr = arr / Max * 255
        arr[arr < 0] = 0
        arr[arr > 255] = 255
    return arr.astype(np.uint8)


def SIFTRegisterTransform(reference: np.ndarray, other: typing.Iterable[np.ndarray], mask: np.ndarray = None, debugPlots: bool = False) -> typing.Tuple[typing.List[np.ndarray], animation.ArtistAnimation]:
    """Given a 2D reference image and a list of other images of the same scene but shifted a bit this function will use OpenCV to calculate the transform from
    each of the other images to the reference. The transforms can be inverted using cv2.invertAffineTransform().
    It will return a list of transforms. Each transform is a 2x3 array in the form returned
    by opencv.estimateAffinePartial2d(). a boolean mask can be used to select which areas will be searched for features to be used
    in calculating the transform. This seems to work much better for normalized images.
    This code is basically a copy of this example, it can probably be improved upon:
    https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_feature2d/py_feature_homography/py_feature_homography.html

    Args:
        reference (np.ndarray): The 2d reference image.
        other (Iterable[np.ndarray]): An iterable containing the images that you want to calculate the translations for.
        mask (np.ndarray): A boolean array indicating which parts of the reference image should be analyzed. If `None` then the whole image will be used.
        debugPlots (bool): Indicates if extra plots should be openend showing the process of the function.

    Returns:
        tuple: A tuple containing:
            List[np.ndarray]:  Returns a list of transforms. Each transform is a 2x3 array in the form returned by opencv.estimateAffinePartial2d(). Note that even
            though they are returned as affine transforms they will only contain translation information, no scaling, shear, or rotation.

            ArtistAnimation: A reference the animation used to diplay the results of the function.
        """
    import cv2

    refImg = to8bit(reference)
    if mask:
        mask = mask.astype(np.uint8)

    #Set up flann matcher
    FLANN_INDEX_KDTREE = 0
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    # Initiate SIFT detector
    sift = cv2.SIFT_create()
    kp1, des1 = sift.detectAndCompute(refImg, mask=mask)

    transforms = []
    if debugPlots:
        anFig, anAx = plt.subplots()
        anims = []
    for i, img in enumerate(other):
        logger.debug(f"Calculating SIFT matches for image {i} of {len(other)}")
        otherImg = to8bit(img)
        # find the keypoints and descriptors with SIFT
        kp2, des2 = sift.detectAndCompute(otherImg, mask=None)

        good = _knnMatch(flann, des1, des2)

        MIN_MATCH_COUNT = 5
        if len(good) > MIN_MATCH_COUNT:
            src_pts = np.float32([kp1[m.queryIdx].pt for m in good]).reshape(-1, 1, 2)
            dst_pts = np.float32([kp2[m.trainIdx].pt for m in good]).reshape(-1, 1, 2)
            M, inlierMask = cv2.estimateAffinePartial2D(src_pts, dst_pts)
            matchesMask = inlierMask.ravel().tolist()
        else:
            logger.warning(f"Image {i}: Not enough matches are found - {len(good)}/{MIN_MATCH_COUNT}")
            M = None
            matchesMask = None
        transforms.append(M)
        if debugPlots and (M is not None):
            anims.append([anAx.imshow(cv2.warpAffine(otherImg, cv2.invertAffineTransform(M), otherImg.shape), 'gray')])
            h, w = refImg.shape
            pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(-1, 1, 2)
            dst = cv2.transform(pts, M)
            draw_params = dict(matchColor=(0, 255, 0),  # draw matches in green color
                               singlePointColor=None,
                               matchesMask=matchesMask,  # draw only inliers
                               flags=2)
            img2 = cv2.polylines(otherImg, [np.int32(dst)], True, 255, 3, cv2.LINE_AA)
            img3 = cv2.drawMatches(refImg, kp1, img2, kp2, good, None, **draw_params)
            fig, ax = plt.subplots()
            ax.imshow(img3, 'gray')
    if debugPlots:
        anFig.suptitle("SIFT: If transforms worked, image should not appear to move.")
        an = animation.ArtistAnimation(anFig, anims)
    else:
        an = None
    return transforms, an


def ORBRegisterTransform(reference: np.ndarray, other: typing.Iterable[np.ndarray], mask: np.ndarray = None, debugPlots: bool = False) -> typing.Tuple[typing.List[np.ndarray], animation.ArtistAnimation]:
    """Given a 2D reference image and a list of other images of the same scene but shifted a bit this function will use OpenCV to calculate the transform from
    each of the other images to the reference. The transforms can be inverted using cv2.invertAffineTransform().
    It will return a list of transforms. Each transform is a 2x3 array in the form returned
    by opencv.estimateAffinePartial2d(). a boolean mask can be used to select which areas will be searched for features to be used
    in calculating the transform. 

    Args:
        reference (np.ndarray): The 2d reference image.
        other (Iterable[np.ndarray]): An iterable containing the images that you want to calculate the translations for.
        mask (np.ndarray): A boolean array indicating which parts of the reference image should be analyzed. If `None` then the whole image will be used.
        debugPlots (bool): Indicates if extra plots should be openend showing the process of the function.

    Returns:
        tuple: A tuple containing:
            List[np.ndarray]:  Returns a list of transforms. Each transform is a 2x3 array in the form returned by opencv.estimateAffinePartial2d(). Note that even
            though they are returned as affine transforms they will only contain translation information, no scaling, shear, or rotation.

            ArtistAnimation: A reference the animation used to diplay the results of the function.
        """
    import cv2

    refImg = to8bit(reference)
    if mask:
        mask = mask.astype(np.uint8)

    #Create FLANN matcher
    FLANN_INDEX_LSH = 6
    index_params = dict(algorithm=FLANN_INDEX_LSH,
                        table_number=6,  # 12
                        key_size=12,  # 20
                        multi_probe_level=1)  # 2
    search_params = dict(checks=100)
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    # Initiate ORB detector
    orb = cv2.ORB_create()
    kp1, des1 = orb.detectAndCompute(refImg, mask=mask)

    transforms = []
    if debugPlots:
        anFig, anAx = plt.subplots()
        anims = []
    for i, img in enumerate(other):
        otherImg = to8bit(img)
        # find the keypoints and descriptors with ORB
        kp2, des2 = orb.detectAndCompute(otherImg, mask=None)

        good = _knnMatch(flann, des1, des2)

        MIN_MATCH_COUNT = 5
        if len(good) > MIN_MATCH_COUNT:
            src_pts = np.float32([kp1[m.queryIdx].pt for m in good]).reshape(-1, 1, 2)
            dst_pts = np.float32([kp2[m.trainIdx].pt for m in good]).reshape(-1, 1, 2)
            M, inlierMask = cv2.estimateAffinePartial2D(src_pts, dst_pts)
            matchesMask = inlierMask.ravel().tolist()
        else:
            logging.getLogger(__name__).warning(f"Image {i}: Not enough matches are found - {len(good)}/{MIN_MATCH_COUNT}")
            M = None
            matchesMask = None
        transforms.append(M)
        if debugPlots and (M is not None):
            anims.append([anAx.imshow(cv2.warpAffine(otherImg, cv2.invertAffineTransform(M), otherImg.shape), 'gray')])
            h, w = refImg.shape
            pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(-1, 1, 2)
            dst = cv2.transform(pts, M)
            draw_params = dict(matchColor=(0, 255, 0),  # draw matches in green color
                               singlePointColor=None,
                               matchesMask=matchesMask,  # draw only inliers
                               flags=2)
            img2 = cv2.polylines(otherImg, [np.int32(dst)], True, 255, 3, cv2.LINE_AA)
            img3 = cv2.drawMatches(refImg, kp1, img2, kp2, good, None, **draw_params)
            fig, ax = plt.subplots()
            ax.imshow(img3, 'gray')
    if debugPlots:
        anFig.suptitle("ORB: If transforms worked, image should not appear to move.")
        an = animation.ArtistAnimation(anFig, anims)
    else:
        an = None
    return transforms, an


def _knnMatch(flannMatcher: cv2.FlannBasedMatcher, des1: np.ndarray, des2: np.ndarray) -> typing.List[cv2.DMatch]:
    """
    Return a list of sufficiently good matches between keypoint descriptors.

    Args:
        flannMatcher: The opencv FLANN matcher object
        des1: An array of `query descriptors` to find matches for.
        des2: An array of `train descriptors` to check as matches for des1
    Returns:
        A list of the selecte opencv DMatch objects
    """
    matches = flannMatcher.knnMatch(des1, des2, k=2)  # Return the two best matches. Note, might only return 1 match
    # store all the good matches as per Lowe's ratio test.
    good = []
    for iMatches in matches:
        if len(iMatches) == 1:
            good.append(iMatches[0])  # If only one match was found keep it
        elif len(iMatches) == 2:
            if iMatches[0].distance < 0.7 * iMatches[1].distance:  # This is known as Lowe's Ratio Test. If two matches were returned only keep the best match if it is sufficiently better than the second best match
                good.append(iMatches[0])
        else:
            raise Exception("Programming Error!")
    return good


def edgeDetectRegisterTranslation(reference: np.ndarray, other: typing.Iterable[np.ndarray], mask: np.ndarray = None, debugPlots: bool = False, sigma: float = 3) -> typing.Tuple[typing.Iterable[np.ndarray], typing.List]:
    """This function is used to find the relative translation between a reference image and a list of other similar images. Unlike `SIFRegisterTransforms` this function
    will not work for images that are rotated relative to the reference. However, it does provide more robust performance for images that do not look identical.

    Args:
        reference (np.ndarray): The 2d reference image.
        other (Iterable[np.ndarray]): An iterable containing the images that you want to calculate the translations for.
        mask (np.ndarray): A boolean array indicating which parts of the reference image should be analyzed. If `None` then the whole image will be used.
        debugPlots (bool): Indicates if extra plots should be openend showing the process of the function.
        sigma (float): this parameter is passed to skimage.feature.canny to detect edges.

    Returns:
        tuple: A tuple containing:
            list[np.ndarray]:  Returns a list of transforms. Each transform is a 2x3 array in the form returned by opencv.estimateAffinePartial2d(). Note that even
            though they are returned as affine transforms they will only contain translation information, no scaling, shear, or rotation.

            list: A list of references to the plotting widgets used to display the results of the function.
    """
    import cv2
    from mpl_qt_viz.visualizers import MultiPlot
    refEd = feature.canny(reference, sigma=sigma)
    if mask is not None: refEd[~mask] = False  # Clear any detected edges outside of the mask
    imEd = [feature.canny(im, sigma=sigma) for im in other]
    affineTransforms = []
    if debugPlots:
        anEdFig, anEdAx = plt.subplots()
        anEdFig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        anEdAx.get_xaxis().set_visible(False)
        anEdAx.get_yaxis().set_visible(False)
        anFig, anAx = plt.subplots()
        anFig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        anAx.get_xaxis().set_visible(False)
        anAx.get_yaxis().set_visible(False)
        anims = [[anAx.imshow(to8bit(reference), 'gray'), anAx.text(100, 100, "Reference", color='r')]]
        animsEd = [[anEdAx.imshow(to8bit(refEd), 'gray'), anEdAx.text(100, 100, "Reference",  color='w')]]
    for i, (im, edgeIm) in enumerate(zip(other, imEd)):
        shifts, error, phasediff = feature.register_translation(edgeIm, refEd)
        logging.getLogger(__name__).info(f"Translation: {shifts}, RMS Error: {error}, Phase Difference:{phasediff}")
        shifts = np.array([[1, 0, shifts[1]],
                           [0, 1, shifts[0]]], dtype=float) # Convert the shift to an affine transform
        affineTransforms.append(shifts)
        if debugPlots:
            plt.figure()
            animsEd.append([anEdAx.imshow(cv2.warpAffine(to8bit(edgeIm), cv2.invertAffineTransform(shifts), edgeIm.shape), 'gray'),  anEdAx.text(100, 100, str(i),  color='w')])
            anims.append([anAx.imshow(cv2.warpAffine(to8bit(im), cv2.invertAffineTransform(shifts), im.shape), 'gray'),  anAx.text(100, 100, str(i), color='r')])
    if debugPlots:
        an = [MultiPlot(anims, "If transforms worked, cells should not appear to move."), MultiPlot(animsEd, "If transforms worked, cells should not appear to move.")]
        [i.show() for i in an]
    else:
        an = []
    return affineTransforms, an


def crossCorrelateRegisterTranslation(reference: np.ndarray, other: typing.Iterable[np.ndarray], debugPlots: bool = False) -> typing.Tuple[typing.Iterable[np.ndarray], 'mpl_qt_viz.visualizers.MultiPlot']:
    """This function is used to find the relative translation between a reference image and a list of other similar images. Unlike `SIFRegisterTransforms` this function
    will not work for images that are rotated relative to the reference.

    Args:
        reference (np.ndarray): The 2d reference image.
        other (Iterable[np.ndarray]): An iterable containing the images that you want to calculate the translations for.
        debugPlots (bool): Indicates if extra plots should be openend showing the process of the function.

    Returns:
        tuple: A tuple containing:
            list[np.ndarray]:  Returns a list of transforms. Each transform is a 2x3 array in the form returned by opencv.estimateAffinePartial2d(). Note that even
                though they are returned as affine transforms they will only contain translation information, no scaling, shear, or rotation.

            MultiPlot: A reference to the plotting widgets used to display the results of the function. If `debugPlots` is False this will be `None`
    """
    import cv2
    if debugPlots:
        anFig, anAx = plt.subplots()
        anFig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        anAx.get_xaxis().set_visible(False)
        anAx.get_yaxis().set_visible(False)
        anims = [[anAx.imshow(to8bit(reference), 'gray'), anAx.text(100, 100, "Reference", color='r')]]
    affineTransforms = []
    for i, im in enumerate(other):
        shifts, error, phasediff = registration.phase_cross_correlation(im, reference, return_error=True)
        logging.getLogger(__name__).debug(f"Translation: {shifts}, RMS Error: {error}, Phase Difference:{phasediff}")
        shifts = np.array([[1, 0, shifts[1]],
                           [0, 1, shifts[0]]], dtype=float) # Convert the shift to an affine transform
        affineTransforms.append(shifts)
        if debugPlots:
            anims.append([
                anAx.imshow(cv2.warpAffine(to8bit(im), cv2.invertAffineTransform(shifts), im.shape), 'gray'),
                anAx.text(100, 100, str(i), color='r')])
    if debugPlots:
        from mpl_qt_viz.visualizers import MultiPlot
        an = MultiPlot(anims, "If transforms worked, images should not appear to move.")
        an.show()
    else:
        an = None
    return affineTransforms, an



from __future__ import annotations
import json
import typing as t_
import os
from pwspy.dataTypes import Acquisition
from pwspy.utility.acquisition import steps


class SequencerCoordinate:
    """
    A coordinate that fully defines a position within a `tree` of steps.

    Args:
        coordSteps: A sequence of tuples of the form (stepId, stepIteration) where `stepId` is the id number of the step being referred to.
            If the step is an iterable step (multiple position, timeseries, etc.) then `stepIteration` should indicate the iteration number,
            otherwise it should be `None`.
        uuid: A universally unique ID string associated with the run of the sequencer that this coordinate is associated with.
    """
    def __init__(self, coordSteps: t_.Sequence[t_.Tuple[int, int]], uuid: str):
        self._fullPath = tuple(coordSteps)
        self.uuid = uuid  # Matches the uuid of the sequence file that ran this acquisition.

    def __repr__(self):
        return f"SeqCoord:{self._fullPath}"

    @staticmethod
    def fromDict(d: dict) -> SequencerCoordinate:
        c = []
        for ID, iteration in zip(d['treeIdPath'], d["stepIterations"]):
            c.append((ID, iteration))
        if 'uuid' in d:
            uuid = d['uuid']
        else:
            uuid = None
        return SequencerCoordinate(c, uuid)

    @staticmethod
    def fromJsonFile(path: str) -> SequencerCoordinate:
        with open(path) as f:
            return SequencerCoordinate.fromDict(json.load(f))

    def isSubPathOf(self, other: SequencerCoordinate):
        """Check if `self` is a parent path of the `item` coordinate """
        assert isinstance(other, SequencerCoordinate)
        if len(self._fullPath) >= len(other._fullPath):
            return False
        return self._fullPath == other._fullPath[:len(self._fullPath)]

    @property
    def iterations(self) -> t_.Sequence[int]:
        return tuple(iteration for ID, iteration in self._fullPath)

    @property
    def ids(self) -> t_.Sequence[int]:
        return tuple(ID for ID, iteration in self._fullPath)

    def getStepIteration(self, step: t_.Union[int, steps.SequencerStep]) -> t_.Optional[int]:
        """

        Args:
            step: May be the ID number of the step or a reference to the actual step.

        Returns:
            The iteration of `Step` that this coordinate corresponds to. If the step is not an iterable step then `None` will be returned.
        """
        if isinstance(step, steps.SequencerStep):
            ID = step.id
        elif isinstance(step, int):
            ID = step
        else:
            raise TypeError(f"`step` arument must be `int` or `SequencerStep`, not `{type(step)}`")
        return [iteration for coordID, iteration in self._fullPath if ID == coordID][0]

    def __eq__(self, other: SequencerCoordinate):
        """Check if these coordinates are identical"""
        assert isinstance(other, SequencerCoordinate)
        return self._fullPath == other._fullPath


class _IterationRangeCoordStep:
    """
    Represents a coordinate for a single step that accepts multiple iterations

    Args:
        ID: The id number of the step being referred to.
        iterations: A sequence of iteration numbers which are considered in the range of iterations this object contains. If `None` then all iterations are accepted.
    """
    def __init__(self, ID: int, iterations: t_.Sequence[int] = None):
        self.stepId = ID
        self.iterations = iterations  #Only iterable step types will have this, most types will keep this as None

    def __contains__(self, item: t_.Tuple[int, int]):
        """
        Args:
            item: A tuple of form (stepId, iteration). See the documentation for SequencerCoordinate
        """
        if self.stepId == item[0]:
            if self.iterations is None:  # This step doesn't have any iterations so there is no need to check anything.
                return True
            elif len(self.iterations) == 0:  # If the accepted iterations are empty then we accept any iteration
                return True
            elif item[1] in self.iterations:
                return True
        return False


class SequencerCoordinateRange:
    """
    A coordinate that can have multiple iterations selected at once.

    Args:
        coordSteps: A sequence of tuples where each tuple represents an acceptable coordinate range for each step in the tree path.
            Tuple is of the form (ID, iterations) where ID is an integer referring to the id number of the step and iterations indicates
            the iterations of that step that are considered in range. `iterations` can be `None` or a sequence of `ints` that are considered in range.
            If `iterations is `None` then all iterations are considered in range. `None` is also the only acceptable value for steps which do not iterate.
    """
    def __init__(self, coordSteps: t_.Sequence[t_.Tuple[int, t_.Optional[t_.Sequence[int]]]]):
        self._fullPath = tuple([_IterationRangeCoordStep(ID, iterations) for ID, iterations in coordSteps])

    def __contains__(self, item: SequencerCoordinate):
        """Returns True if this is a subpath of `item` and the iteration at each step lies within the range of acceptable iterations for this object"""
        if not isinstance(item, SequencerCoordinate):
            return False
        for i, coordRange in enumerate(self._fullPath):
            if not (item._fullPath[i] in coordRange):
                return False
        return True

    def setAcceptedIterations(self, stepId: int, iterations: t_.Optional[t_.Sequence[int]]):
        """Sets the acceptable iterations for the step associated with `stepId`

        Args:
            stepId: The id of the step you want to adjust the iteration range for
            iterations: A sequence if iteration numbers to include. If `None` then all iterations are accepted.
        """
        for coordStep in self._fullPath:
            if coordStep.stepId == stepId:
                coordStep.iterations = iterations
                return
        raise ValueError(f"No step with ID: {stepId} was found.")


class SequenceAcquisition:
    """
    An object linking and acquisition with a sequencerCoordinate file.

    Attributes:
        acquisition: The pwspy.dataTypes.Acquisition object linking to the raw data
        sequencerCoordinate: The coordinate object locating the acquisition within a sequence.
    """
    def __init__(self, acquisition: Acquisition):
        self.acquisition = acquisition
        path = os.path.join(acquisition.filePath, "sequencerCoords.json")
        self.sequencerCoordinate = SequencerCoordinate.fromJsonFile(path)  # This will throw an exception if no sequence coord is found.






# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Using the wave transfer matrix formalism from chapter 7 of Saleh and Teich
Fundamentals of Photonics, this script calculates the reflectance of a
multilayer dielectric.
http://www.phys.ubbcluj.ro/~emil.vinteler/nanofotonica/TemeControl_FCMD014_Vinteler.pdf
https://en.wikipedia.org/wiki/Transfer-matrix_method_(optics)

m is the final transfer matrix. It should be made by multiplying the matrices
representing each element of the system. If the transmitted light is considered
to be propagating from left to right then the matrices should be in multiplied
in reverse, from right to left.


Classes
----------
.. autosummary::
   :toctree: generated/

   Polarization
   Layer
   Stack
   NonPolarizedStack

"""
__all__ = ['Polarization', 'Layer', 'Stack', 'NonPolarizedStack']

import typing
from enum import Enum, auto

import matplotlib.pyplot as plt
from cycler import cycler
from numbers import Number
from typing import Union, Optional, List
from pwspy.utility.reflection import Material
import pandas as pd
import numpy as np
import matplotlib as mpl


class Polarization(Enum):
    """An enumeration of the possible polarization types."""
    TE = auto()  # Transverse Electric Field
    TM = auto()  # Transverse Magnetic Field


class Layer:
    """This represents a layer with a thickness and an index of refraction. Note: This whole system only supports
    lossless media, we only use the real part of the index of refraction.

    Args:
        mat: This can either be a number or series of numbers representing the refractive index at different wavelengths
            or it can be a `Material` in which case the refractive index will be automatically calculated.
        d: The thickness of the layer. The units that thicknesses and wavelengths are specified in must match.
        name: An optional name which will be dislayed if the layer is plotted
    """
    def __init__(self, mat: Union[Number, pd.Series, Material], d: float, name: typing.Optional[str] = None):
        if not isinstance(mat, (Number, pd.Series, Material)):
            raise TypeError(f"Type {type(mat)} is not supported")
        self.mat = mat
        self.d = d
        if name is None and isinstance(mat, Material):
            self.name = mat.name
        else:
            self.name = name

    def getRefractiveIndex(self, wavelengths: np.ndarray) -> pd.Series:
        """Get the refractive index of the layer.

        Args:
            wavelengths: The wavelengths to calculate the refractive index at.

        Returns:
            The refractive index.
        """
        from .reflectanceHelper import getRefractiveIndex
        if isinstance(self.mat, Material):
            n = getRefractiveIndex(self.mat, wavelengths=wavelengths)
            n = pd.Series(np.real(n), index=n.index)
            return n
        elif isinstance(self.mat, Number):
            return pd.Series(np.array([self.mat]*len(wavelengths)), index=wavelengths)
        elif isinstance(self.mat, pd.Series):
            return self.mat


class StackBase:
    def __init__(self, wavelengths: Union[Number, np.ndarray], elements: Optional[List[Layer]] = None):
        assert len(wavelengths.shape) == 1
        self.wavelengths = wavelengths
        if elements is None:
            self.layers = []
        else:
            self.layers = elements

    def addLayer(self, element: Layer):
        """Add a new `Layer` to the `Stack`

        Args:
            element: A new layer to add.
        """
        self.layers.append(element)


class NonPolarizedStack(StackBase):
    """Represents a stack of 1d homogenous films. Reflectance can only be calculated at 0 incidence angle in which case
    polarization is irrelevant. This class does not do anything that can't be done with `Stack`. Indices of refraction must
    be real (no absorption).

    Args:
        wavelengths: The wavelengths that calculation should operate over.
        elements: The initial layers to add to the stack.

    """

    def _generateMatrix(self) -> np.ndarray:
        """First and last items just have propagation matrices."""
        matrices = []
        lastItem: Layer = None
        for el in self.layers:
            if lastItem is not None:
                matrices.append(self.interfaceMatrix(lastItem.getRefractiveIndex(self.wavelengths), el.getRefractiveIndex(self.wavelengths)))
            matrices.append(self.propagationMatrix(el.getRefractiveIndex(self.wavelengths), el.d))
            lastItem = el
        matrices.reverse()

        previousMat = None
        for matrix in matrices:
            if previousMat is not None:
                previousMat = previousMat @ matrix  # Matrix multiplication
            else:
                previousMat = matrix
        return previousMat

    def plot(self):
        """Open a Matplotlib plot of the stack."""
        cycle = cycler('color', ['r', 'g', 'b', 'y', 'c', 'm'])
        fig, ax = plt.subplots()
        ax.set_prop_cycle(cycle)
        startCoord = 0
        for el, col in zip(self.layers, cycle()):
            r = plt.Rectangle((startCoord, 0), el.d, 1, color=col['color'])
            t = plt.Text(r.xy[0]+r.get_width()/2, .5, f"{el.name}: {np.mean(el.getRefractiveIndex(self.wavelengths))}")
            startCoord = startCoord + el.d
            ax.add_patch(r)
            ax.add_artist(t)
        ax.set_xlim([0, startCoord])
        plt.show()


    @staticmethod
    def interfaceMatrix(n1: pd.Series, n2: pd.Series) -> np.ndarray:
        """Generate a matrix representing the interface between two dielectrics with indices n1 on the left and n2 on the right.
        Actually the order of terms does not appear to matter. This does not account for polarization or incidence angles
        other than 0 degrees.

        Args:
            n1: The refractive indices on one side of the reflective interface
            n2: The refractive indices on the other side of the reflective interface

        Returns:
            A transfer matrix for the reflective interface
        """
        assert len(n1) == len(n2)
        assert np.all(n1.index == n2.index)
        n1 = np.array(n1)
        n2 = np.array(n2)
        matrix = np.array([[n2 + n1, n2 - n1],
                           [n2 - n1, n2 + n1]])
        matrix = np.transpose(matrix, axes=(2, 0, 1))
        matrix = (1 / (2 * n2))[:, None, None] * matrix
        assert matrix.shape == (len(n1),) + (2, 2)
        return matrix


    @staticmethod
    def propagationMatrix(n: pd.Series, d: float) -> np.ndarray:
        """Returns a matrix representing the propagation of light. n should be a pandas Series where the values are
        complex refractive index and the index fo the Series is the associated wavelengths. with wavelength.
        for a distance of `d`. d and the wavelengths must use the same units.

        Args:
            n: The refractive indices of the material
            d: The distance of propagation.

        Returns:
              A transfer matrix for propagation through a material.

        """
        wavelengths = n.index
        phi = np.array(2 * np.pi * d * n / wavelengths)
        zeroArray = 0 * phi  # Without this our matrix will not shape properly
        matrix = np.array([[np.exp(-1j * phi), zeroArray],
                           [zeroArray, np.exp(1j * phi)]])
        matrix = np.transpose(matrix, axes=(2, 0, 1))
        assert matrix.shape == (len(n),) + (2, 2)
        return matrix

    def calculateReflectance(self) -> np.ndarray:
        """Calculate the reflectance for this `Stack`.

        Returns:
            The reflectance.
        """
        m = self._generateMatrix()
        assert m.shape == (len(self.wavelengths),) + (2, 2)
        scatterMatrix = np.array([ # A 2x2 scattering matrix. https://en.wikipedia.org/wiki/S-matrix
            [m[:, 0, 0] * m[:, 1, 1] - m[:, 0, 1] * m[:, 1, 0], m[:, 0, 1]],
            [-m[:, 1, 0],                              np.ones((m.shape[0],))]])
        scatterMatrix = np.transpose(scatterMatrix, axes=(2, 0, 1))
        scatterMatrix = (1 / m[:, 1, 1])[:, None, None] * scatterMatrix
        R = scatterMatrix[:, 1, 0] * np.conjugate(scatterMatrix[:, 1, 0])  # The reflectance of the stack. This is a real number. Equivalent to np.absolute(scatterMatrix[1, 0]) ** 2
        assert np.max(np.abs(np.imag(R))) == 0
        return np.real(R)


class Stack(StackBase):
    """Represents a stack of 1d homogenous films. Reflectance for the two polarizations can be calculated for a range of
    numerical apertures (angles). Indices of refraction must be real (no absorption).

    Args:
        wavelengths: The wavelengths that calculation should operate over.
        elements: The initial layers to add to the stack.

    """
    @staticmethod
    def interfaceMatrix(n1: pd.Series, n2: pd.Series, polarization: Polarization, NAs: np.ndarray) -> np.ndarray:
        """Returns a matrix representing a dieletric interface. n1 and n2 should be a pandas Series where the values are
        complex refractive index and the index of the Series is the associated wavelengths.

        Args:
            n1: The refractive indices on one side of the reflective interface
            n2: The refractive indices on the other side of the reflective interface
            polarization: The polarization that should be used for the calculation
            NAs: An array of the numerical aperture values. :todo: More details would be good

        Returns:
            A transfer matrix for the reflective interface

        """
        assert len(n1) == len(n2)
        assert np.all(n1.index == n2.index)
        n1 = n1.values[:, None]
        n2 = n2.values[:, None]
        NAs = NAs[None, :]
        theta1 = np.arcsin(NAs / n1)
        theta2 = np.arcsin(NAs / n2)
        if polarization is Polarization.TE:
            a21 = 1
            N1 = n1 * np.cos(theta1)
            N2 = n2 * np.cos(theta2)
        else:  # Polarization.TM
            a21 = np.cos(theta2) / np.cos(theta1)
            N1 = n1 / np.cos(theta1)
            N2 = n2 / np.cos(theta2)
        matrix = np.array([[N2 + N1, N2 - N1],
                           [N2 - N1, N2 + N1]])
        matrix = np.transpose(matrix, axes=(2, 3, 0, 1))
        matrix = (1 / (2 * a21 * N2))[:, :, None, None] * matrix
        assert matrix.shape == (n1.size, NAs.size) + (2, 2)
        return matrix


    @staticmethod
    def propagationMatrix(n: pd.Series, d: float, NAs: np.ndarray) -> np.ndarray:
        """Returns a matrix representing the propagation of light for a distance of `d`.
        d and the wavelengths must use the same units.

        Args:
            n: The refractive indices of the material
            d: The distance of propagation.

        Returns:
              A transfer matrix for propagation through a material.

        """
        wavelengths = np.array(n.index)[:, None]
        n = n.values[:, None]
        NAs = NAs[None, :]
        theta = np.arcsin(NAs / n)
        phi = n * d * np.cos(theta) * 2 * np.pi / wavelengths
        # phi = np.array(2 * np.pi * d * n / wavelengths)
        zeroArray = np.zeros(phi.shape)  # Without this our matrix will not shape properly
        matrix = np.array([[np.exp(-1j * phi), zeroArray],
                           [zeroArray, np.exp(1j * phi)]])
        matrix = np.transpose(matrix, axes=(2, 3, 0, 1)) # Move the angle and wavelength dimensions up front. leave the matrix dimensions as the last two.
        assert matrix.shape == (n.size, NAs.size) + (2, 2)
        return matrix

    def _generateMatrix(self, polarization: Polarization, NAs: np.ndarray) -> np.ndarray:
        """First and last items just have propagation matrices. """
        matrices = []
        lastItem: Layer = None
        for el in self.layers:
            if lastItem is not None:
                matrices.append(self.interfaceMatrix(lastItem.getRefractiveIndex(self.wavelengths), el.getRefractiveIndex(self.wavelengths), polarization, NAs))
            matrices.append(self.propagationMatrix(el.getRefractiveIndex(self.wavelengths), el.d, NAs))
            lastItem = el
        matrices.reverse()

        previousMat = None
        for matrix in matrices:
            if previousMat is not None:
                previousMat = previousMat @ matrix  # Matrix multiplication
            else:
                previousMat = matrix
        return previousMat

    def calculateReflectance(self, NAs: np.ndarray):
        """Given an array of numerical apertures this function returns the reflectance as a dictionary of 2d arrays.
        There is one 2d array for each of the two polarizations. the dimensions of the array is (wavelengths x NAs).
        The total reflectance can be calculated as the average reflectance of the two polarizations. Other ellipsometric
        parameters can also be calculated.

        Args:
            NAs: The numerical apertures to calculate reflectance at.

        Returns:
            A dictionary containing a reflectance array for each of the two polarizations. The polarization is the key to the dictionary.
            Each reflectance is a MxN array where M is the number of wavelengths and N is the number of NAs passed to this function.
        """
        out = {}
        for polarization in Polarization:
            m = self._generateMatrix(polarization, NAs)
            scatterMatrix = np.array([  # A 2x2 scattering matrix. https://en.wikipedia.org/wiki/S-matrix
                [m[:, :, 0, 0] * m[:, :, 1, 1] - m[:, :, 0, 1] * m[:, :, 1, 0], m[:, :, 0, 1]],
                [-m[:, :, 1, 0],                              np.ones((m.shape[0], m.shape[1]))]])
            scatterMatrix = np.transpose(scatterMatrix, axes=(2, 3, 0, 1))
            scatterMatrix = (1 / m[:, :, 1, 1])[:, :, None, None] * scatterMatrix
            R = scatterMatrix[:, :, 1, 0] * np.conjugate(scatterMatrix[:, :, 1, 0])  # The reflectance of the stack. This is a real number. Equivalent to np.absolute(scatterMatrix[1, 0]) ** 2
            out[polarization] = R.real
        return out

    def circularIntegration(self, NAs: np.ndarray) -> pd.Series:
        """Given an array of NumericalApertures (usually from 0 to NAMax.) This function integrates the reflectance over
        a disc of Numerical Apertures (Just like in a microscope the Aperture plane is a disc shape, with higher NA
        being further from the center.) Ultimately the result of this integration should match the reflectance measured
        with the same NA.

        Args:
            NAs: The numerical apertures to calculate reflectance at.

        Returns:
            A pandas `Series` with wavelengths as the index and reflectance as the value.
        """
        from scipy.integrate import trapz
        d = self.calculateReflectance(NAs)
        r = (d[Polarization.TE] + d[Polarization.TM]) / 2  # Reflectance averaged over polarization.
        r = r * 2 * np.pi * NAs #The aperture plane is a disk, higher NAs have larger circumference disks contributing to them. hence the 2pi*na factor.
        inte = trapz(r, NAs, axis=1)
        int2 = trapz(2 * np.pi * NAs, NAs) #This is used for normalization. basically accounting for the fact that na is proportional to radius in aperture plane, not equal.
        return pd.Series(inte / int2, index=self.wavelengths)

    def plot(self, NAs: np.ndarray, polarization: Polarization = None):
        """Plot various graphs of reflectance vs NA. NAs should be an array of Numerical apertures to have the
        reflectance calculated for. `polarization` can be specified to view the reflectance of only one polarization."""
        d = self.calculateReflectance(nas)
        rTM = d[Polarization.TM]
        rTE = d[Polarization.TE]
        if polarization is None:
            eccentricity = np.sqrt(1 - rTM / rTE)        # rTM == a**2, rTE == b**2
            r = (rTM + rTE) / 2
        elif polarization == Polarization.TE:
            r = rTE
            eccentricity = None
        elif polarization == Polarization.TM:
            r = rTM
            eccentricity = None
        else:
            raise TypeError(f"`polarization` must be Polarization, not {type(polarization)}.")
        fig2, ax2 = plt.subplots()
        fig2.suptitle("R")
        ax2.set_ylabel("wavelength")
        ax2.set_xlabel("NA")
        if eccentricity is not None:
            fig, ax = plt.subplots()
            fig.suptitle("Eccentricity")
            ax.set_ylabel("wavelength")
            ax.set_xlabel("Angle")
            im = ax.imshow(eccentricity, extent=[NAs[0], NAs[-1], self.wavelengths[-1], self.wavelengths[0]], interpolation=None, aspect='auto')
            plt.colorbar(im, ax=ax)
            fig.show()
        im = ax2.imshow(r, extent=[NAs[0], NAs[-1], self.wavelengths[-1], self.wavelengths[0]], interpolation=None, aspect='auto')
        plt.colorbar(im, ax=ax2)
        fig2.show()
        fig3, ax3 = plt.subplots()
        colormap = plt.cm.gist_rainbow
        colors = [colormap(i) for i in np.linspace(0, 0.99, r.shape[1])]
        norm = mpl.colors.Normalize(vmin=NAs[0], vmax=NAs[-1])
        sm = plt.cm.ScalarMappable(cmap=colormap, norm=norm)
        for i in range(r.shape[1]):
            ax3.plot(self.wavelengths, r[:, i], color=colors[i], label=NAs[i])
        plt.colorbar(sm, ax=ax3)
        fig3.show()
        fig4, ax4 = plt.subplots()
        ax4.set_xlabel("NA")
        ax4.plot(NAs, rTM.mean(axis=0), label='TM')
        ax4.plot(NAs, rTE.mean(axis=0), label='TE')
        ax4.plot(NAs, r.mean(axis=0), label='R')
        ax4.legend()
        fig4.show()


if __name__ == '__main__':
    num = 40
    wv = np.linspace(500, 700, num=100)
    s = Stack(wv)
    s.addLayer(Layer(Material.Glass, 100000))
    for i in range(20):
        s.addLayer(Layer(1.2, 650/1.2/4))
        s.addLayer(Layer(1.35, 650/1.35/4))
    # s.addLayer(Layer(Material.ITO, 2000))
    s.addLayer(Layer(Material.Air, 10000))

    nas = np.linspace(0, .2, num=1000)
    s.plot(nas)
    out = s.circularIntegration(nas)
    fig, ax = plt.subplots()
    ax.plot(s.wavelengths, out)
    fig.show()
    pass



# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import typing
import collections.abc

NumberOrArray = typing.Union[typing.Sequence[float], float]


class S2DMatlabBridge:  # TODO implement the SigmaToDApprox method as well.
    """
    Opens a MATLAB process to run the Sigma2D conversion code. You must have the MATLAB engine for Python installed
    in your Python environment.

    Args:
        s2dPath: The file path to the SigmaConversion MATLAB package.
    """
    def __init__(self, s2dPath: str):
        import matlab.engine
        self._engine = matlab.engine.start_matlab()
        self._engine.addpath(s2dPath)

    def createRIDefinitionFromGladstoneDale(self, mediaRI: float, CVC: float) -> 'matlab.object':
        """
        Create an S2D.RIDefinition object from the GladstoneDale equation.

        Args:
            mediaRI: The refractive index of the media that the chromatin is immersed in.
            CVC: The CVC, also referred to as Phi. The ratio of Chromatin Volume : Total Volume, kind of like density.

        Returns:
            A S2D.RIDefinition object.
        """
        return self._engine.S2D.RIDefinition.createFromGladstoneDale(mediaRI, CVC)

    def createSystemConfiguration(self, ri_def: 'S2D.RIDefinition', na_i: float, na_c: float, center_lambda: float,
                                  oil_immersion: bool, cell_glass_interface: bool) -> 'matlab.object':
        """
        Create a system configuration object. This object contains all information about the microscope and sample refractive indices.

        Args:
            ri_def: An RI definition object
            na_i: The illumination NA of the objective.
            na_c: The collection NA of the objective.
            center_lambda: The center wavelength of illumination. This should be the center in K-space. So, the wavelength of the center wavenumber.
            oil_immersion: If True then the objective is treated as though it is immersed in RI_glass. If False it is treated as though it is immersed in RI_media.
            cell_glass_interface: If True then the cell/glass interface is treated as the reference reflection of the configuration. If False then the cell/media interface is treated as the reference reflection.
        Returns:
            A S2D.SystemConfiguration MATLAB object.
        """
        return self._engine.S2D.SystemConfiguration(ri_def, na_i, na_c, center_lambda, oil_immersion, cell_glass_interface)

    def SigmaToD_AllInputs(self, sigmaIn: NumberOrArray, system_config: 'S2D.SystemConfiguration',
                           Nf: float, thickIn: float) -> NumberOrArray:
        """
        Run the complete sigma to D conversion function.

        Args:
            sigmaIn: The sigma values you want to convert.
            system_config: The SystemConfiguration object to use.
            Nf: The genomic length of a packing domain.
            thickIn: The expected thickness of the sample.

        Returns:
            dOut: This is analogous to `D_b` in Aya's paper. The `model parameter`.
            dCorrected: This is the true `D`. This is usually what we care about.
            Nf_expected: The genomic length we expect based on D and the calculated lMax.
            lmax_corrected: Calculation of LMax from Nf and Db based on eqn. 2.
        """
        if isinstance(sigmaIn, collections.abc.Sequence):
            import matlab
            sigmaIn = matlab.double(sigmaIn)
        return self._engine.SigmaToD_AllInputs(sigmaIn, system_config, Nf, thickIn, nargout=4)

    def close(self):
        """
        Close the MATLAB engine. This may not be necessary as the engine will be automatically closed when Python shuts down.
        """
        self._engine.quit()


if __name__ == '__main__':
    s = S2DMatlabBridge(r'C:\Users\backman05\Documents\Bitbucket\SigmaToD\SigmaConversion')
    ri = s.createRIDefinitionFromGladstoneDale(1.337, .3)
    conf = s.createSystemConfiguration(ri, .52, 1.49, 585, True, True)
    out = s.SigmaToD_AllInputs([.1, .1, .2], conf, 1e6, 3)
    a = 1



import pytest
import shapely.geometry
import numpy as np
import pwspy.dataTypes as pwsdt


def test_roi(sequenceData):
    """Test that ROIs can be detected, loaded, and that the various required attributes have the expected data types."""
    acqs = [pwsdt.Acquisition(i) for i in sequenceData.datasetPath.glob("Cell[0-9]")]
    for acq in acqs:
        roiSpecs = acq.getRois()
        for roiSpec in roiSpecs:
            roiFile = acq.loadRoi(*roiSpec)
            assert roiFile.acquisition is acq
            roi = roiFile.getRoi()
            assert isinstance(roi.polygon, shapely.geometry.Polygon)
            assert isinstance(roi.verts, np.ndarray)
            assert len(roi.verts.shape) == 2
            assert roi.verts.shape[1] == 2



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
import json
import os.path as osp
import h5py
import numpy as np
import typing as t_
import pandas as pd

from pwspy import __version__ as pwspyversion
from pwspy.analysis.warnings import AnalysisWarning
from pwspy.utility.fileIO import processParallel
from pwspy.utility.misc import cached_property
if t_.TYPE_CHECKING:
    from pwspy.dataTypes import ICBase, MetaDataBase


class AbstractAnalysisSettings(ABC):
    """This abstract class lays out the basic skeleton of what an AnalysisSettings class should implement.
    These classes represent everything about the settings of an anlysis that can be reliably saved and then loaded again.
    The settings that are actually passed to the analyiss are the RuntimeSettings which can contain items that can't reliably be
    loaded from a json file (for example, references to data files which may not still be at the same file path if someone tried to load the settings"""

    @classmethod
    def fromJson(cls, filePath: str, name: str) -> AbstractAnalysisSettings:
        """Create a new instance of this class from a json text file.

        Args:
            filePath: The path to the folder containing the JSON file to be loaded.
            name: The name that the analysis was saved as.
        Returns:
            A new instance of an analysis settings class.
        """
        with open(osp.join(filePath, f'{name}_{cls.FileSuffix}.json'), 'r') as f:
            d = json.load(f)
        return cls._fromDict(d)

    def toJson(self, filePath: str, name: str):
        """Save this object to a json text file.

        Args:
            filePath: The path to the folder to contain the new JSON file.
            name: The name to save the analysis as.
        """
        d = self.asDict()
        with open(osp.join(filePath, f'{name}_{self.FileSuffix}.json'), 'w') as f:
            json.dump(d, f, indent=4)

    def toJsonString(self) -> str:
        """Use `_asDict` to convert an instance of this class to a json string.

        Returns:
            A JSON formatted string.
        """
        return json.dumps(self.asDict(), indent=4)

    @classmethod
    def fromJsonString(cls, string: str) -> AbstractAnalysisSettings:
        """Use `_fromDict` to load a new instance of the `cls` from a json string.

        Args:
            string: A JSON formatted string to load the object from.
        Returns:
            A new instance of analysis settings class.
        """
        return cls._fromDict(json.loads(string))

    @abstractmethod
    def asDict(self) -> dict:
        """

         Returns:
            A dictionary with setting names as the keys and the values of the settings as the values.
        """
        pass

    @classmethod
    @abstractmethod
    def _fromDict(cls, d: dict) -> AbstractAnalysisSettings:
        """
        Args:
            d: A dictionary containing analysis settings.
        Returns:
            A new instance of the analysis settings class based on `d`."""
        pass

    @property
    @abstractmethod
    def FileSuffix(self):
        """A constant string which will be used as a suffix for json files saving the settings of this class."""
        pass


class AbstractAnalysis(ABC):
    """This abstract class lays out the basic skeleton that an analysis class should implement."""
    @abstractmethod
    def __init__(self):
        """Does all of the one-time tasks needed to start running an analysis. e.g. prepare the reference, load the extrareflection cube, etc."""
        pass

    @abstractmethod
    def run(self, cube: ICBase) -> t_.Tuple[AbstractAnalysisResults, t_.List[AnalysisWarning]]:
        """Given an data cube to analyze this function returns an instanse of AnalysisResults. In the PWSAnalysisApp this function is run in parallel by the AnalysisManager.

        Args:
            cube: A data cube to be analyzed using the settings provided in the constructor of this class.
        Returns:
            A new instance of analysis results.
        """
        pass

    @abstractmethod
    def copySharedDataToSharedMemory(self):
        """When running the `run` method in parallel memory for the object used must be copied to each new process. We can avoid that and save a lot of Ram by moving data
        that is shared between processes to shared memory. If you don't want to implement this then just override it and raise NotImplementedError"""
        pass


class AbstractAnalysisResults(ABC):
    """This abstract class lays out the most basic skeleton of what an AnalysisResults object should implement."""

    _currentmoduleversion = pwspyversion  # We save the version of this code that was used to save the results. For retrospection

    @classmethod
    @abstractmethod
    def create(cls) -> AbstractAnalysisResults:
        """Used to create results from existing variables. These results can then be saved to file.

        Returns:
            A new instance of analysis results.
        """
        pass

    @classmethod
    @abstractmethod
    def load(cls, directory: str, name: str):
        """Used to load results from a saved file.

        Args:
            directory: The path to the folder containing the analysis file.
            name: The name that the analysis was saved as.
        Returns:
            A new instance of analysis results loaded from file.
        """
        pass


def _clearError(func):
    """This decorator tries to run the original function. If the function raises a keyerror then we raise a new keyerror with a clearer message. This is intended to be used with `field` accessors of implementations
    of `AbstractHDFAnalysisResults`."""
    def newFunc(*args):
        try:
            return func(*args)
        except KeyError:
            raise KeyError(f"The analysis file does not contain a {func.__name__} item.")
    newFunc.__name__ = func.__name__  # failing to do this renaming can mess with other decorators e.g. cached_property
    newFunc.__doc__ = func.__doc__
    return newFunc


def _getFromDict(func):
    """This decorator makes it so that when the method is run we will check if our class instance has a `file` property. If not, then we will attempt to access a `dict` property which is keyed
    by the same name as our original method. Otherwide we simply run the method. This is intended for use with implementations of `AbstractHDFAnalysisResults`"""
    def newFunc(self, *args):
        if self.file is None:
            return self.dict[func.__name__]
        else:
            return func(self, *args)

    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    return newFunc


class AbstractHDFAnalysisResults(AbstractAnalysisResults):
    """
    This abstract class implements methods of `AbstractAnalysisResults` for an object that can be saved and loaded to/from an HDF file.
    And the `fileName2Name` method which does the opposite operation. Can be instantiated with one of the two constructor arguments.
    The constructor should not be run directly, it should only be used by the `create` and `load` class methods.

    Args:
        file: To load from file provide the `h5py.File`.
        variablesDict: To create a new object from variables, provide a dictionary keyed by all the field names.
        analysisName: Optionally store the name of the analysis.
    """
    @staticmethod
    def FieldDecorator(func):
        """Decorate functions in subclasses that access their fields from the HDF file with this decorator. It will:
        1: Make it so the data is load from disk on the first access and stored in memory for every further access.
        2: Report an understandable error if the field isn't found in the HDF file.
        3: Make the accessors work even if the the object isn't associated with an HDF file."""
        return cached_property(_clearError(_getFromDict(func)))

    #TODO this holds onto the reference to the h5py.File meaning that the file can't be deleted until the object has been deleted. Maybe that's good. but it causes some problems.
    def __init__(self, file: t_.Optional[h5py.File] = None, variablesDict: t_.Optional[dict] = None, analysisName: t_.Optional[str] = None):
        if file is not None:
            assert variablesDict is None
        elif variablesDict is not None:
            assert file is None
        self.file = file
        self.dict = variablesDict
        self.analysisName = analysisName

    @cached_property
    def moduleVersion(self) -> str:
        """The version of PWSpy code that this file was saved with."""
        return bytes(np.array(self.file['pwspy_version'])).decode()

    @staticmethod
    @abstractmethod
    def fields() -> t_.Tuple[str, ...]:
        """

        Returns:
            A sequence of string names of the datafields that the analysis results contains.
        """
        pass

    @staticmethod
    @abstractmethod
    def name2FileName(name: str) -> str:
        """

        Args:
            name: An analysis name.
        Returns:
            The corresponding file name for the hdf5 file.
        """
        pass

    @staticmethod
    @abstractmethod
    def fileName2Name(fileName: str) -> str:
        """

        Args:
            fileName: The filename that the HDF file was saved as.

        Returns:
            The analysis name.
        """
        pass

    def toHDF(self, directory: str, name: str, overwrite: bool = False, compression: str = None):
        """
        Save the AnalysisResults object to an HDF file in `directory`. The name of the file will be determined by `name`. If you want to know what the full file name
        will be you can use this class's `name2FileName` method.

        Args:
            directory: The path to the folder to save the file in.
            name: The name of the analysis. This determines the file name.
            overwrite: If `True` then any existing file of the same name will be replaced.
            compression: The value of this argument will be passed to h5py.create_dataset for numpy arrays. See h5py documentation for available options.
        """
        from pwspy.dataTypes import ICBase  # Need this for instance checking
        fileName = osp.join(directory, self.name2FileName(name))
        if (not overwrite) and osp.exists(fileName):
            raise OSError(f'{fileName} already exists.')
        with open(fileName, 'wb') as pythonFile:
            with h5py.File(pythonFile, 'w', driver='fileobj') as hf:  # Using the default driver causes write errors when writing from windows to a Samba shared server. Using a reference to a python `File Object` solves this issue.
                # Save version
                hf.create_dataset('pwspy_version', data=np.string_(self._currentmoduleversion))
                # Save fields defined by implementing subclass
                for field in self.fields():
                    k = field
                    v = getattr(self, field)
                    if isinstance(v, AbstractAnalysisSettings):
                        v = v.toJsonString() # Convert to string, then string case will then handle saving the string.
                    elif isinstance(v, dict):  # Save as json. The str case will handle the actual saving.
                        v = json.dumps(v)
                    if isinstance(v, str):
                        hf.create_dataset(k, data=np.string_(v))  # h5py recommends encoding strings this way for compatability.
                    elif isinstance(v, ICBase):
                        hf = v.toHdfDataset(hf, k, fixedPointCompression=True)
                    elif isinstance(v, np.ndarray):
                        hf.create_dataset(k, data=v, compression=compression)
                    elif v is None:
                        pass
                    else:
                        raise TypeError(f"Analysis results type {k}, {type(v)} not supported or expected")

    @classmethod
    def load(cls, directory: str, name: str) -> AbstractHDFAnalysisResults:
        """Load an analyis results object from an HDF5 file located in `directory`.

        Args:
            directory: The path to the folder containing the file.
            name: The name of the analysis.
        Returns:
            A new instance of analysis results loaded from file.
        """
        filePath = osp.join(directory, cls.name2FileName(name))
        if not osp.exists(filePath):
            raise OSError(f"The {cls.__name__} analysis file does not exist. {filePath}")
        file = h5py.File(filePath, 'r')
        return cls(file, None, name)

    def __del__(self):
        if self.file is not None:  # Make sure to release the file if it's still open.
            try:
                self.file.close()
            except:  # Sometimes when python is shutting down this causes an error. Doesn't matter though.
                pass




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
from pwspy.utility.micromanager import PropertyMap

from pwspy.utility.micromanager.positions import PositionList

"""This example demonstrates how to use generate new cell positions from a set of positions after the sample has been picked up and likely shifted or rotated.
This method relies on measuring a set (at least 3) of reference positions before and after moving the dish. You can then use these positions to generate an 
affine transform. This affine transform can then be applied to your original cell positions in order to generate a new set of positions for the same cells.
In the case of a standard cell culture dish it is best to use the corners of the glass coverslip as your reference locations.

@author: Nick Anthony
"""

if __name__ == '__main__':

    import pathlib as pl
    import matplotlib.pyplot as plt
    from skimage.transform import AffineTransform
    import numpy as np

    plt.ion()
    NCPath = pl.Path(r'Z:\Nick\NU-NC_EtOH fixation comparison\NC\Buccal cell')
    NUPath = pl.Path(r'Z:\Nick\NU-NC_EtOH fixation comparison\NU\Buccal')

    # Load the position list of the coverslip corners taken at the beginning of the experiment.
    preTreatRefPositions = PositionList.fromNanoMatFile(NCPath / 'corners_list2.mat', 'TIXYDrive')
    # Load the position list of the coverslip corners after placing the dish back on the microscope after treatment.
    postTreatRefPositions = PositionList.fromPropertyMap(PropertyMap.loadFromFile(NUPath / 'corners.pos'))
    # Generate an affine transform describing the difference between the two position lists.
    transformMatrix = preTreatRefPositions.getAffineTransform(postTreatRefPositions)
    # Load the positions of the cells we are measuring before the dish was removed.
    preTreatCellPositions = PositionList.fromNanoMatFile(NCPath / 'cell_list.mat', 'TIXYDrive')
    # Transform the cell positions to the new expected locations.
    postTreatCellPositions = preTreatCellPositions.applyAffineTransform(transformMatrix)
    # Save the new positions to a file that can be loaded by Micro-Manager.
    postTreatCellPositions.toPropertyMap().saveToFile(NUPath / 'transformedPositions.pos')

    # Plot the reference and cell positions before treatment
    fig, ax = plt.subplots()
    preTreatRefPositions.plot(fig, ax)
    preTreatCellPositions.plot(fig, ax)

    # Plot the reference and cell positions after treatment
    fig2, ax2 = plt.subplots()
    postTreatRefPositions.plot(fig2, ax2)
    postTreatCellPositions.plot(fig2, ax2)

    af = AffineTransform(np.vstack([transformMatrix, [0, 0, 1]]))
    print(f"Scale: {af.scale}, Shear: {af.shear}, Rotation: {af.rotation / (2*np.pi) * 360}, Translation: {af.translation}")



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Classes used in the analysis of Dynamics data.

Classes
---------

.. autosummary::
    :toctree: generated/

    DynamicsAnalysisSettings
    DynamicsAnalysisResults
    DynamicsAnalysis

Inheritance
-------------
.. inheritance-diagram:: DynamicsAnalysisSettings DynamicsAnalysisResults DynamicsAnalysis
    :parts: 1

"""

from __future__ import annotations
import dataclasses
import logging
from datetime import datetime

import numpy as np
import pandas as pd
from numpy import ma
import multiprocessing as mp
import typing as t_
from . import AbstractAnalysis, warnings, AbstractAnalysisSettings, AbstractHDFAnalysisResults
from pwspy import dateTimeFormat
import pwspy.dataTypes as pwsdt
from pwspy.utility.reflection import reflectanceHelper, Material

__all__ = ['DynamicsAnalysis', 'DynamicsAnalysisSettings', 'DynamicsAnalysisResults']


class DynamicsAnalysis(AbstractAnalysis):
    """This class performs the analysis of RMS_t_squared and D (diffusion). It is based on a set of MATLAB scripts written
    by Scott Gladstein. The original scripts can be found in the `_oldMatlab` subpackage.

    References:
         "Multimodal interferometric imaging of nanoscale structure and macromolecular motion uncovers UV induced cellular paroxysm"

    Args:
        settings: The settings use for the analysis
        extraReflectance: the metadata object referring to a calibration file for extra reflectance. You can optionally proide the ExtraReflectanceCube rather than just the metadata object referring to it.
        ref: A reference acquisition to use for normalization.
    """
    n_medium = 1.37  # The average index of refraction for chromatin?

    def __init__(self, settings: DynamicsAnalysisSettings, extraReflectance: t_.Optional[t_.Union[pwsdt.ERMetaData, pwsdt.ExtraReflectanceCube]], ref: pwsdt.DynCube):
        super().__init__()
        logger = logging.getLogger(__name__)

        if not ref.processingStatus.cameraCorrected:
            ref.correctCameraEffects(settings.cameraCorrection)
        if not ref.processingStatus.normalizedByExposure:
            ref.normalizeByExposure()
        if ref.metadata.pixelSizeUm is not None:  # Only works if pixel size was saved in the metadata.
            ref.filterDust(.75)  # Apply a blur to filter out dust particles. This is in microns. I'm not sure if this is the optimal value.

        if settings.referenceMaterial is None:
            theoryR = pd.Series(np.ones((len(ref.times),)), index=ref.times)  # Having this as all ones effectively ignores it.
            logger.warning("DynamicsAnalysis ignoring reference material correction")
        else:
            theoryR = reflectanceHelper.getReflectance(settings.referenceMaterial, Material.Glass, wavelengths=ref.metadata.wavelength, NA=settings.numericalAperture)

        if isinstance(extraReflectance, pwsdt.ERMetaData): # In the case the extraReflectance is an ExtraReflectanceCube or `None` no action needs to take place.
            extraReflectance = pwsdt.ExtraReflectanceCube.fromMetadata(extraReflectance)

        if extraReflectance is None:
            Iextra = None 
            logger.warning("DynamicsAnalysis ignoring extra reflection")
        else:
            if extraReflectance.metadata.numericalAperture != settings.numericalAperture:
                logger.warning("The numerical aperture of your analysis does not match the NA of the Extra Reflectance Calibration. Calibration File NA: {extraReflectance.metadata.numericalAperture}. PWSAnalysis NA: {settings.numericalAperture}.")
            idx = np.asarray(np.array(extraReflectance.wavelengths) == ref.metadata.wavelength).nonzero()[0][0]  # The index of extra reflectance that matches the wavelength of our dynamics cube
            I0 = ref.data.mean(axis=2) / (float(theoryR) + extraReflectance.data[:, :, idx])  # I0 is the intensity of the illumination source, reconstructed in units of `counts`. this is an inversion of our assumption that reference = I0*(referenceReflectance + extraReflectance)
            Iextra = I0 * extraReflectance.data[:, :, idx]  # Convert from reflectance to predicted counts/ms.
            ref.subtractExtraReflection(Iextra)  # remove the extra reflection from our data#
        if not settings.relativeUnits:
            ref = ref / theoryR[None, None, :]  # now when we normalize by our reference we will get a result in units of physical reflectance rather than arbitrary units.

        self.refMean = ref.data.mean(axis=2)
        ref.normalizeByReference(self.refMean)  # We normalize so that the average is 1. This is for scaling purposes with the AC. Seems like the AC should be scale independent though, not sure.
        self.refAc = ref.getAutocorrelation()[:, :, :settings.diffusionRegressionLength+1].mean(axis=(0, 1))  # We find the average autocorrlation of the background to cut down on noise, presumably this is uniform accross the field of view any way, right?
        self.refTag = ref.metadata.idTag
        self.erTag = extraReflectance.metadata.idTag if extraReflectance is not None else None

        self.settings = settings
        self.extraReflection = Iextra

    def run(self, cube: pwsdt.DynCube) -> t_.Tuple[DynamicsAnalysisResults, t_.List[warnings.AnalysisWarning]]:  # Inherit docstring
        warns = []
        if not cube.processingStatus.cameraCorrected:
            cube.correctCameraEffects(self.settings.cameraCorrection)
        if not cube.processingStatus.normalizedByExposure:
            cube.normalizeByExposure()
        if self.extraReflection is not None:
            cube.subtractExtraReflection(self.extraReflection)
        cube.normalizeByReference(self.refMean)

        cubeAc = cube.getAutocorrelation()
        cubeAc = cubeAc[:, :, :self.settings.diffusionRegressionLength+1] # We are only going to use the first few time points of the ACF, we can get rid of the rest.
        rms_t_squared = cubeAc[:, :, 0] - self.refAc[0]  # The rms^2 noise of the reference averaged over the whole image.
        rms_t_squared[rms_t_squared < 0] = 0  # Sometimes the above noise subtraction can cause some of our values to be barely below 0, that's going to be a problem.
        # If we didn't care about noise subtraction we could get rms_t as just `cube.data.std(axis=2)`

        # Determine the mean-reflectance for each pixel in the cell.
        reflectance = cube.data.mean(axis=2)

        # Diffusion
        cubeAc = ma.array(cubeAc)  # Convert to the numpy.MaskedArray type to help us mark some data as invalid.
        cubeAc[cubeAc.data[:, :, 0] < np.sqrt(2)*self.refAc[0]] = ma.masked  # Remove pixels with low SNR. Default threshold removes values where 1st point of acf is less than sqrt(2) of background acf
        ac = ma.array(cubeAc - self.refAc)  # Background subtracted autocorrelation function.
        ac = ac / ac[:, :, 0][:, :, None]  # Normalize by the zero-lag value
        ac[np.any(ac <= 0, axis=2)] = ma.masked  # Before taking the log of the autocorrelation any negative or zero values will cause problems. Remove the pixel entirely

        dt = (cube.times[-1] - cube.times[0]) / (len(cube.times) - 1) / 1e3  # Convert to seconds
        k = (self.n_medium * 2 * np.pi) / (cube.metadata.wavelength / 1e3)  # expressing wavelength in microns to match up with old matlab code.
        val = np.log(ac) / (4 * k ** 2)  # See the `theory` section of the paper for an explanation of the 4k^2. The slope of log(ac) should be equivalent to 1/t_c in the paper.
        d_slope = -self._maskedLinearRegression(val, dt)  # Get the slope of the autocorrelation. This is related to the diffusion in the cell. The minus is here to make the number positive, the slope is really negative.

        results = DynamicsAnalysisResults.create(meanReflectance=reflectance,
                                                 rms_t_squared=rms_t_squared,
                                                 reflectance=cube,
                                                 diffusion=d_slope,
                                                 settings=self.settings,
                                                 imCubeIdTag=cube.metadata.idTag,
                                                 referenceIdTag=self.refTag,
                                                 extraReflectionIdTag=self.erTag)

        return results, warns

    @staticmethod
    def _maskedLinearRegression(arr: ma.MaskedArray, dt: float) -> np.ndarray:
        """
        Takes a 3d ACF array as input and returns a 2d array indicating the slope along the 3rd dimension of the input array.
        The dimensions of the output array match the first two dimensions of the input array. The input array can have invalid pixels masked out, this function
        will exclude them from the calculation.

        Args:
            arr: The masked 3D array of the autocorrelation function of each spectra.
            dt: The time interval between each element of the autocorrelation function.
        Returns:
            The 2D array containing the slope of each ACF at each pixel.
        """
        origShape = arr.shape
        y = np.reshape(arr, (origShape[0]*origShape[1], origShape[2]))  #Convert to a 2d array [pixels, time]. This is required by the polyfit function.
        t = np.array([i*dt for i in range(origShape[2])]) # Generate a 1d array representing the time axis.
        # Y = np.reshape(y[~y.mask], (np.sum((~y.mask).sum()//y.shape[1], y.shape[1])))  # Remove all the masked pixels since the polyfit function doesn't know what to do with them.
        Y = np.reshape(y[~y.mask], ((~y.mask).sum()//y.shape[1], y.shape[1]))  # Remove all the masked pixels since the polyfit function doesn't know what to do with them.
        coeff = np.polyfit(t, Y.T, deg=1)  # Linear fit
        slope = coeff[0, :]  # Get the slope and ignore the y intercept
        Slope = ma.zeros(y.shape[0]) # Create an empty masked array that includes all pixels again
        Slope.mask = y.mask[:, 0] # Copy the original mask indicating which pixels are invalid.
        Slope[~Slope.mask] = slope  # Fill our calculated slopes into the yunmasked pixels.
        Slope = np.reshape(Slope, (origShape[0], origShape[1])) # Reshape back to a 2d image
        return Slope

    def copySharedDataToSharedMemory(self): # Inherit docstring
        refdata = mp.RawArray('f', self.refAc.size)
        refdata = np.frombuffer(refdata, dtype=np.float32).reshape(self.refAc.shape)
        np.copyto(refdata, self.refAc)
        self.refAc = refdata

        refmdata = mp.RawArray('f', self.refMean.size)
        refmdata = np.frombuffer(refmdata, dtype=np.float32).reshape(self.refMean.shape)
        np.copyto(refmdata, self.refMean)
        self.refMean = refmdata

        if self.extraReflection is not None:
            iedata = mp.RawArray('f', self.extraReflection.size)
            iedata = np.frombuffer(iedata, dtype=np.float32).reshape(self.extraReflection.shape)
            np.copyto(iedata, self.extraReflection)
            self.extraReflection = iedata


class DynamicsAnalysisResults(AbstractHDFAnalysisResults): # Inherit docstring.
    @staticmethod
    def fields():   # Inherit docstring.
        return ('meanReflectance', 'reflectance', 'rms_t_squared', 'diffusion', 'time',
                'settings', 'imCubeIdTag', 'referenceIdTag', 'extraReflectionIdTag')

    @staticmethod
    def name2FileName(name: str) -> str: # Inherit docstring.
        return f'dynAnalysisResults_{name}.h5'

    @staticmethod
    def fileName2Name(fileName: str) -> str: # Inherit docstring.
        return fileName.split('dynAnalysisResults_')[1][:-3]

    @classmethod
    def create(cls, settings: DynamicsAnalysisSettings, meanReflectance: np.ndarray, rms_t_squared: np.ndarray, reflectance: pwsdt.DynCube, diffusion: np.ndarray,
               imCubeIdTag: str, referenceIdTag: str, extraReflectionIdTag: t_.Optional[str]): # Inherit docstring.
        d = {'time': datetime.now().strftime(dateTimeFormat),
            'meanReflectance': meanReflectance,
            'reflectance': reflectance,
            'diffusion': diffusion,
            'rms_t_squared': rms_t_squared,
            'imCubeIdTag': imCubeIdTag,
            'referenceIdTag': referenceIdTag,
            'extraReflectionIdTag': extraReflectionIdTag,
            'settings': settings}
        return cls(None, d)

    @AbstractHDFAnalysisResults.FieldDecorator
    def meanReflectance(self) -> np.ndarray:
        """A 2D array giving the reflectance of the image averaged over the full spectra."""
        dset = self.file['meanReflectance']
        return np.array(dset)

    @AbstractHDFAnalysisResults.FieldDecorator
    def rms_t_squared(self) -> np.ndarray:
        """A 2D array giving the spectral variance at each position in the image."""
        dset = self.file['rms_t_squared']
        return np.array(dset)

    @AbstractHDFAnalysisResults.FieldDecorator
    def settings(self) -> DynamicsAnalysisSettings:
        """The settings used to generate these results."""
        return DynamicsAnalysisSettings.fromJsonString(bytes(np.array(self.file['settings'])).decode())

    @AbstractHDFAnalysisResults.FieldDecorator
    def reflectance(self) -> pwsdt.DynCube:
        """A dynamics cube containing the 3D reflectance array after all corrections and analysis."""
        dset = self.file['reflectance']
        return pwsdt.DynCube.fromHdfDataset(dset)

    @AbstractHDFAnalysisResults.FieldDecorator
    def diffusion(self) -> np.ndarray:
        """A 2D array indicating the diffusion at each position in the image."""
        dset = self.file['diffusion']
        return np.array(dset)

    @AbstractHDFAnalysisResults.FieldDecorator
    def imCubeIdTag(self) -> str:
        """The idtag of the dynamics cube that was analyzed."""
        return bytes(np.array(self.file['imCubeIdTag'])).decode()

    @AbstractHDFAnalysisResults.FieldDecorator
    def referenceIdTag(self) -> str:
        """The idtag of the dynamics cube that was used as a reference for normalization."""
        return bytes(np.array(self.file['referenceIdTag'])).decode()

    @AbstractHDFAnalysisResults.FieldDecorator
    def time(self) -> str:
        """The time that the analysis was performed."""
        return self.file['time']

    @AbstractHDFAnalysisResults.FieldDecorator
    def extraReflectionIdTag(self) -> str:
        """The idtag of the extra reflection correction that was used."""
        return bytes(np.array(self.file['extraReflectionIdTag'])).decode()


@dataclasses.dataclass
class DynamicsAnalysisSettings(AbstractAnalysisSettings):
    """These settings determine the behavior of the `DynamicsAnalysis` class.

        Args:
            extraReflectanceId: The unique `IDTag` of the extraReflectance calibration that was used on this analysis.
            referenceMaterial: The material that was imaged in the reference image of this analysis. Found as an in pwspy.moduleConst.Material. The
                theoretically predicted
                reflectance of the reference image is used in the extraReflectance correction.
            numericalAperture: The illumination NA of the system. This is used for two purposes. First, we want to make sure that the NA of our data matches
                the NA of our extra reflectance correction cube.
                Second, the theoretically predicted reflectance of our reference is based not only on what our refereMaterial is but also the NA since
                reflectance is angle dependent.
            relativeUnits: If `True` then all calculation are performed such that the reflectance is 1 if it matches the reference. If `False` then we use the
                theoretical reflectance of the reference  (based on NA and reference material) to normalize our results to the actual physical reflectance of
                the sample (about 0.4% for water)
            cameraCorrection: An object describing the dark counts and non-linearity of the camera used. If the data supplied to the DynamicsAnalysis class has already been corrected then this
                setting will not be used. Setting this to `None` will result in the camera correcting being automatically determined based on the image files' metadata.
            diffusionRegressionLength: The original matlab scripts for analysis of dynamics data determined the slope of the log(ACF) by looking only at the
                first two indices, (log(ACF)[1]-log(ACF)[0])/dt. This results in very noisy results. However as you at higher index value of the log(ACF) the
                noise becomes much worse. A middle ground is to perform linear regression on the first 4 indices to determine the slope. You can adjust that
                number here.
    """
    extraReflectanceId: t_.Optional[str]
    referenceMaterial: Material
    numericalAperture: float
    relativeUnits: bool
    cameraCorrection: t_.Optional[pwsdt.CameraCorrection]
    diffusionRegressionLength: int = 3

    FileSuffix = "dynAnalysis"  # This is used for setting the filename when saving and loading to json

    def __post_init__(self):
        assert self.diffusionRegressionLength > 0
        assert self.diffusionRegressionLength < 20  # Even 20 is probably way too long, unless a system is created with extremely low noise.

    def asDict(self) -> dict:
        d = dataclasses.asdict(self)
        if self.referenceMaterial is None:
            d['referenceMaterial'] = None
        else:
            d['referenceMaterial'] = self.referenceMaterial.name  # Convert from enum to string
        return d

    @classmethod
    def _fromDict(cls, d: dict) -> DynamicsAnalysisSettings:
        if d['referenceMaterial'] is not None:
            d['referenceMaterial'] = Material[d['referenceMaterial']]  # Convert from string to enum
        for newKey in ['cameraCorrection']:
            if newKey not in d.keys():
                d[newKey] = None  # For a while these settings were missing from the code. Allow us to still load old files.
        return cls(**d)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Functions which check data during analysis for abnormality. If abnormal conditions are found
then an AnalysisWarning is produced. The application can then display or record these warnings. This aspect of the program has not really
been fully implemented.
"""

from typing import Optional
import numpy as np
#TODO actually use these!


class AnalysisWarning:
    """Each warning should have a short description and a longer explanation.

    Args:
        shortMsg: A brief explanation of the warning.
        longMsg: A longer in depth explanation.
    """
    def __init__(self, shortMsg: str, longMsg: str):
        self.longMsg = longMsg
        self.shortMsg = shortMsg


def checkMeanSpectraRatio(ratio: float) -> Optional[AnalysisWarning]:
    """For PWS while we expect RMS of the spectra in a region to be similar we don't expect nearby pixels to have well
    correlated spectra. By averaging the spectra of all pixels in a region we can get a `mean spectra`. Ideally we would
    expect this to be flat but realistically we se some variation due to the natural spectral reflectance profile of the sample.
    If this mean spectra changes it suggests something has changed in the system and we want to be aware of this."""
    if ratio > 0.4:  # TODO the upper and lower bounds here were arbitrarily chosen without any testing.
        return AnalysisWarning("Mean RMS ratio too high", f"Ratio between variance of mean ROI spectra and mean of spectra variance in ROI is {ratio} (>0.4). This suggests that the ROI is absorbing or fluorescing, or something stranger could be happening.")
    elif ratio < 0.3:
        return AnalysisWarning("Mean RMS ratio too low", f"Ratio between variance of mean ROI spectra and mean of spectra variance in ROI is {ratio} (<0.3). This suggests that the ROI is absorbing or fluorescing, or something stranger could be happening.")
    else:
        return None


def checkRSquared(rSquared: np.ndarray) -> Optional[AnalysisWarning]:
    """We try to fit a line to the natural log of the autocorrelation function in order to find the slope. The RSquared
    tells us how well this fit worked. If RSquared is too low then we can't really trust the slop value that we calculated."""
    if np.any(rSquared < 0.7):  # TODO this threshold was arbitrarily chosen.
        return AnalysisWarning("R^2 too low", f"{np.sum(rSquared<0.7)} of the {rSquared.size} elements in our ROI show this problem. The rSquared of our linear fit to the autocorrelation function is (<0.7) the LD and CDR parameters may not not be valid.")
    else:
        return None




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from dataclasses import dataclass
import typing as t_
import numpy as np

from ._abstract import AbstractCompilerSettings, AbstractRoiCompilationResults, AbstractRoiCompiler
from .. import warnings
from ...dataTypes import Roi
from ...dataTypes._other import RoiFile

if t_.TYPE_CHECKING:
    from ..pws import PWSAnalysisResults

# TODO Documentation


@dataclass
class PWSCompilerSettings(AbstractCompilerSettings):
    """These settings determine which values should be processed during compilation"""
    reflectance: bool = False
    rms: bool = False
    polynomialRms: bool = False
    autoCorrelationSlope: bool = False
    rSquared: bool = False
    ld: bool = False
    opd: bool = False
    meanSigmaRatio: bool = False


@dataclass
class PWSRoiCompilationResults(AbstractRoiCompilationResults):
        cellIdTag: str
        analysisName: str
        reflectance: float
        rms: float
        polynomialRms: float
        autoCorrelationSlope: float
        rSquared: float
        ld: float
        opd: np.ndarray
        opdIndex: np.ndarray  # The x axis of a plot of opd
        varRatio: float #The ratio of signal variance of the Roi's mean spectra to the mean signal variance (rms^2) of the roiFile. should be between 0 and 1.


class PWSRoiCompiler(AbstractRoiCompiler):
    settings: PWSCompilerSettings  # Just doing this to get accurate type hinting

    def __init__(self, settings: PWSCompilerSettings):
        super().__init__(settings)

    def run(self, results: PWSAnalysisResults, roi: Roi) -> t_.Tuple[PWSRoiCompilationResults, t_.List[warnings.AnalysisWarning]]:
        warns = []
        reflectance = self._avgOverRoi(roi, results.meanReflectance) if self.settings.reflectance else None
        rms = self._avgOverRoi(roi, results.rms) if self.settings.rms else None
        if self.settings.polynomialRms:
            try:
                polynomialRms = self._avgOverRoi(roi, results.polynomialRms)
            except KeyError:
                polynomialRms = None
        else:
            polynomialRms = None

        if self.settings.autoCorrelationSlope:
            try:
                autoCorrelationSlope = self._avgOverRoi(roi, results.autoCorrelationSlope,
                    condition=np.logical_and(results.rSquared > 0.9,
                        results.autoCorrelationSlope < 0))
            except KeyError:
                autoCorrelationSlope = None
        else:
            autoCorrelationSlope = None

        if self.settings.rSquared:
            try:
                warns.append(warnings.checkRSquared(results.rSquared[roi.mask]))
                rSquared = self._avgOverRoi(roi, results.rSquared)
            except KeyError:
                rSquared = None
        else:
            rSquared = None

        if self.settings.ld:
            try:
                ld = self._avgOverRoi(roi, results.ld)
            except KeyError:
                ld = None
        else:
            ld = None

        if self.settings.opd:
            try:
                opd, opdIndex = results.opd
                opd = opd[roi.mask].mean(axis=0)
            except KeyError:
                opd = opdIndex = None
        else:
            opd = opdIndex = None

        if self.settings.meanSigmaRatio:
            try:
                spectra = results.reflectance.getMeanSpectra(roi)[0]
                meanRms = spectra.std()
                varRatio = meanRms**2 / (results.rms[roi.mask] ** 2).mean()
                warns.append(warnings.checkMeanSpectraRatio(varRatio))
            except KeyError:
                varRatio = None
        else:
            varRatio = None

        results = PWSRoiCompilationResults(
                    cellIdTag=results.imCubeIdTag,
                    analysisName=results.analysisName,
                    reflectance=reflectance,
                    rms=rms,
                    polynomialRms=polynomialRms,
                    autoCorrelationSlope=autoCorrelationSlope,
                    rSquared=rSquared,
                    ld=ld,
                    opd=opd,
                    opdIndex=opdIndex,
                    varRatio=varRatio)
        warns = [w for w in warns if w is not None]  # Strip None from warns list
        return results, warns

    @staticmethod
    def _avgOverRoi(roi: Roi, arr: np.ndarray, condition: np.ndarray = None) -> float:
        """Returns the average of arr over the ROI.
        if condition is provided then only value of arr where the condition is satisfied are included."""
        assert len(arr.shape) == 2
        if condition is not None:
            return arr[np.logical_and(roi.mask, condition)].mean()
        else:
            return arr[roi.mask].mean()



# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This script saves pws analysis results to the test data. This must be run before many of the other examples will work.
"""

from pwspy import analysis
from pwspy import dataTypes as pwsdt
from examples import PWSExperimentPath

settings = analysis.pws.PWSAnalysisSettings.loadDefaultSettings("Recommended")

# Load our blank reference image which will be used for normalization.
refAcq = pwsdt.Acquisition(PWSExperimentPath / 'Cell3')
ref = refAcq.pws.toDataClass()

anls = analysis.pws.PWSAnalysis(settings=settings, extraReflectance=None, ref=ref)  # Create a new analysis for the given reference image and analysis settings. The "ExtraReflection" calibration is ignored in this case.

acq = pwsdt.Acquisition(PWSExperimentPath / "Cell1")  # Create an "Acquisition" object to handle operations for the data associated with a single acquisition
cube = acq.pws.toDataClass()  # Request that the PWS metadata object load the full data.
results, warnings = anls.run(cube)  # Run the pre-setup analysis on our data. Get the analysis results and potentially a list of warnings from the analysis.

acq.pws.saveAnalysis(results, 'myAnalysis', overwrite=True)  # Save our analysis results to file in the default location alongside the raw data under the `analyses` folder.

loadedResults = acq.pws.loadAnalysis('myAnalysis')  # This is just going to be a copy of `results`, loaded from file.



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
import numpy as np
from pwspy.dataTypes import Roi, Acquisition
import matplotlib.pyplot as plt
import matplotlib


def roiColor(data: np.ndarray, rois: List[Roi], vmin, vmax, scale_bg, hue=0, exponent=1, numScaleBarPix=0):
    """
    Given a 2D image of data this function will scale the data, apply an exponential curve, and color the ROI regions with Hue.
    Used in many presentations and publications.

    Args:
        data (np.ndarray): an MxN array of data to be imaged
        rois (List[Roi]): a list of Roi objects. the regions inside a roiFile will be colored.
        vmin (float): the minimum value in the data that will be set to black
        vmax (float): the maximum value in the data that will be set to white
        scale_bg (float): Scales the brightness of the background (non-roiFile) region.
        hue (float): A value of 0-1 indicating the hue of the colored regions.
        exponent (float): The exponent used to curve the color map for more pleasing results.
        numScaleBarPix (float): The length of the scale bar in number of pixels.

    Returns:
        np.ndarray: MxNx3 RGB array of the image"""

    mask = np.zeros(rois[0].mask.shape, dtype=np.bool)
    for roi in rois:
        mask = np.logical_or(mask, roi.mask)

    # scale and process rms cube (this is probably not the best way to do it)
    data = data - vmin
    data[data < 0] = 0
    data[data > (vmax - vmin)] = vmax - vmin
    data = data ** exponent
    data = data * 1 / ((vmax - vmin) ** exponent) # normalize image so maximum value is 1

    # make the nucs red and everything else gray scale
    hsv = np.ones((data.shape[0], data.shape[1], 3))
    hsv[:,:,2] = data
    hsv[:,:,2][~mask] *= scale_bg
    hsv[:,:,0] = hue
    hsv[:,:,1][~mask] = 0
    out = matplotlib.colors.hsv_to_rgb(hsv)

    if (numScaleBarPix > 0):
        out[round(out.shape[0]*.965):round(out.shape[0]*.975), round(out.shape[0]*.03):round(out.shape[0] * .03 + numScaleBarPix), :] = 1

    return out


if __name__ == "__main__":
    acq = Acquisition(r'G:\Aya_NAstudy\matchedNAi_largeNAc\cells\Cell3')
    an = acq.pws.loadAnalysis('p0')
    rois = [acq.loadRoi(name, num).getRoi() for name, num, fformat in acq.getRois() if name == 'nucleus']
    out = roiColor(an.rms, rois, 0.04, .3, 1, exponent=1.1, numScaleBarPix=100)
    fig, ax = plt.subplots()
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    ax.imshow(out)#, vmax=vmax, vmin=vmin)
    fig.show()
    a = 1



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""A package containing functionality useful for calculation reflections.

Subpackages
-------------

.. autosummary::
    :toctree: generated/

    extraReflectance
    multilayerReflectanceEngine
    reflectanceHelper

Classes
---------

.. autosummary::
   :toctree: generated/

   Material
"""

import enum


@enum.unique
class Material(enum.Enum):
    """
    An enumeration class containing items for the various materials that we can calculate reflectance for.
    """
    Glass = enum.auto()
    Water = enum.auto()
    Air = enum.auto()
    Silicon = enum.auto()
    Oil_1_7 = enum.auto()
    Oil_1_4 = enum.auto()
    Ipa = enum.auto()
    Ethanol = enum.auto()
    ITO = enum.auto()
    Methanol = enum.auto()
    Glycerol_99_5 = enum.auto()  # 99.5 purity Glycerol



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import typing
from dataclasses import dataclass

import numpy as np

from ._abstract import AbstractCompilerSettings, AbstractRoiCompilationResults
from ...dataTypes._other import RoiFile


@dataclass
class GenericCompilerSettings(AbstractCompilerSettings):
    """These settings determine which values should be processed during compilation"""
    roiArea: bool = False


@dataclass
class GenericRoiCompilationResults(AbstractRoiCompilationResults):
        """Results for compilation that don't pertain to any specific analysis."""
        roiFile: RoiFile
        roiArea: int  # the number of pixels of an ROI


class GenericRoiCompiler:
    def __init__(self, settings: GenericCompilerSettings):
        self.settings = settings

    def run(self, roi: RoiFile) -> GenericRoiCompilationResults:
        if self.settings.roiArea:
            roiArea: typing.Optional[int] = np.sum(roi.getRoi().mask)
        else:
            roiArea = None

        results = GenericRoiCompilationResults(
                    roiFile=roi,
                    roiArea=roiArea)
        return results




# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Functions related to fluorescent images.

Functions
-----------
.. autosummary::
   :toctree: generated/

   updateFolderStructure

"""

import os
from glob import glob
import numpy as np
import tifffile as tf
import pwspy.dataTypes as pwsdt


def updateFolderStructure(rootDirectory: str, rotate: int, flipX: bool, flipY: bool):
    """Used to translate old fluorescence images to the new file organization that is recognized by the code.

    Args:
        rootDirectory: The top level directory containing fluorescence images that were saved in the old `FL_Cell{X}` folder format
        rotate: The number of times that the images should be rotated clockwise to match up with the PWS images they go with
        flipX: Should the images be mirrored over the X-axis after being rotated?
        flipY: Should the images be mirrored over the Y-axis after being rotated?

    """

    files = glob(os.path.join(rootDirectory, '**', 'FL_Cell*'), recursive=True)
    for file in files:
        cellNum = int(file.split('FL_Cell')[-1])
        parentPath = file.split("FL_Cell")[0]
        data = tf.imread(os.path.join(file, 'image_bd.tif'))
        data = np.rot90(data, k=rotate)
        if flipX:
            data = np.flip(data, axis=1)
        if flipY:
            data = np.flip(data, axis=0)
        fl = pwsdt.FluorescenceImage(data, {'exposure': None})
        newPath = os.path.join(parentPath, f'Cell{cellNum}', 'Fluorescence')
        os.mkdir(newPath)
        fl.toTiff(newPath)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This package includes functionality for loading experiment information saved by the PWS "Acquisition Sequencer" which is part of the PWS plugin
for Micro-Manager.

Functions
----------
.. autosummary::
   :toctree: generated/

   loadDirectory

Classes
---------
.. autosummary::
   :toctree: generated/

   RuntimeSequenceSettings
   SequenceAcquisition
   SequencerCoordinate
   SequencerCoordinateRange
   SequencerStep
   IterableSequencerStep
   ZStackStep
   TimeStep
   PositionsStep
   ContainerStep

Inheritance
-------------
.. inheritance-diagram:: TreeItem SequencerStep IterableSequencerStep ZStackStep TimeStep PositionsStep ContainerStep RuntimeSequenceSettings SequenceAcquisition SequencerCoordinate SequencerCoordinateRange
    :parts: 1

"""

import typing as t_
import warnings

from pwspy.utility.acquisition.steps import RuntimeSequenceSettings

from .steps import SequencerStep, IterableSequencerStep, ZStackStep, TimeStep, PositionsStep, ContainerStep
from ._treeItem import TreeItem
from .sequencerCoordinate import SequenceAcquisition, SequencerCoordinate, SequencerCoordinateRange
import os
import pwspy.dataTypes as pwsdt
from glob import glob


def loadDirectory(directory: os.PathLike) -> t_.Tuple[SequencerStep, t_.List[SequenceAcquisition]]:
    """
    If `directory` contains a dataset acquired with the acquisition sequencer then this function will return a python
    object representing the sequence settings and a list of references to the acquisitions that are part of the sequence.

    Args:
        directory: The file path to the dataset directory.

    Returns:
        A tuple containing:
            The Root `SequencerStep` of the acquisition sequence.
            A list of `SequenceAcquisition` objects belonging to the sequence.
    """
    rtSeq = RuntimeSequenceSettings.fromJsonFile(directory)
    if rtSeq.uuid is None:
        warnings.warn("Old acquisition sequence file must have been loaded. No UUID found. Acquisitions returned by this function may not actually belong to this sequence.")

    files = glob(os.path.join(directory, '**', 'Cell[0-9]*'), recursive=True)
    acqs = []
    for f in files:
        try:
            acqs.append(SequenceAcquisition(pwsdt.Acquisition(f)))
        except FileNotFoundError:
            pass  # There may be "Cell" folders that don't contain a sequencer coordinate.
    acqs = [acq for acq in acqs if acq.sequencerCoordinate.uuid == rtSeq.uuid]  # Filter out acquisitions that don't have a matching UUID to the sequence file.
    # TODO verify that all expected acquisitions were found.
    return rtSeq.rootStep, acqs



# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Loop through all ROIs for all acquisitions in a directory and plot a histogram of the RMS values within the ROI.
"""

import pwspy.dataTypes as pwsdt
import matplotlib.pyplot as plt
import pathlib
import numpy as np
from examples import PWSExperimentPath
plt.ion()

workingDirectory = PWSExperimentPath  # The folder that all your acquisitions are saved under.
analysisName = 'script'  # This will often be "p0"

def plotHist(roi, rms):
    """
    This function takes an ROI
    and a 2D RMS image and plots a histogram of the RMS values inside the ROI
    """
    # Check input values just to be safe.
    assert isinstance(roi, pwsdt.Roi)  # Make sure roiFile variable is actually an ROI
    assert isinstance(rms, np.ndarray)  # Make sure the RMS image is an numpy array
    assert roi.mask.shape == rms.shape  # Make sure the ROI and RMS arrays have the same dimensions.

    vals = rms[roi.mask]  # A 1D array of the values inside the ROI
    plt.hist(vals)  # Plot a histogram


cellFolderIterator = pathlib.Path(workingDirectory).glob("Cell[0-9]")  # An iterator for all folders that are below workingDirectory and match the "regex" pattern "Cell[0-9]"
for folder in cellFolderIterator:
    acq = pwsdt.Acquisition(folder)  # An object handling the contents of a single "Cell{X}" folder

    try:
        anls = acq.pws.loadAnalysis(analysisName)  # Load the analysis results from file.
    except:
        print(f"Analysis loading failed for {acq.filePath}")
        continue  # Skip to the next loop iteration

    roiSpecs = acq.getRois()  # A list of the names, numbers, and fileFormats of the ROIs in this acquisition

    for name, number, fformat in roiSpecs:  # Loop through every ROI.
        roiFile = acq.loadRoi(name, number, fformat)  # Load the ROI from file.
        plotHist(roiFile.getRoi(), anls.rms)  # Use the function defined above to plot a histogram



# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
This script blurs an image cube in the xy direction. Allows you to turn an
image of cells into something that can be used as a reference image, assuming
most of the the FOV is glass. In reality you should just have a good reference image to use and not resort to something
like this.
"""

import copy

import matplotlib.pyplot as plt
import pwspy.dataTypes as pwsdt
from examples import PWSImagePath

plt.ion()

acq = pwsdt.Acquisition(PWSImagePath)
a = acq.pws.toDataClass()

a.correctCameraEffects()  # Correct for dark counts and potentially for camera nonlinearity using metadata stored with the original measurement.
a.normalizeByExposure()  # Divide by exposure time to get data in units of `counts/ms`. This isn't strictly necessary in this case.

mirror = copy.deepcopy(a)
mirror.filterDust(10)  # Apply a gaussian blurring with sigma=10 microns along the XY plane.

a.plotMean()  # Plot the mean reflectance of the original
mirror.plotMean()  # Plot the mean reflectance after filtering.
a.normalizeByReference(mirror)  # Normalize raw by reference
a.plotMean()  # Plot the measurement after normalization.
plt.figure()
plt.imshow(a.data.std(axis=2))  # Plot RMS after normalization.
